from unittest.mock import Mock
from authtuna.helpers import (
    get_remote_address,
    get_device_data,
    is_username_valid,
    is_password_valid,
    sanitize_username,
)
import pytest


@pytest.mark.asyncio
async def test_get_remote_address():
    """Tests for various IP address headers."""
    # Test with Cloudflare header
    request = Mock(headers={"CF-Connecting-IP": "192.168.1.1"}, client=None)
    assert await get_remote_address(request) == "192.168.1.1"

    # Test with X-Forwarded-For header
    request = Mock(headers={"X-Forwarded-For": "203.0.113.1"}, client=None)
    assert await get_remote_address(request, other_ip_headers=["X-Forwarded-For"]) == "203.0.113.1"

    # Test fallback to client host
    request = Mock(headers={}, client=Mock(host="127.0.0.1"))
    assert await get_remote_address(request) == "127.0.0.1"

    # Test default IP
    request = Mock(headers={}, client=None)
    assert await get_remote_address(request) == "127.0.0.1"

@pytest.mark.asyncio
async def test_get_device_data():
    """Test device data parsing."""
    headers = {
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36",
        "CF-IPCity": "Mountain View",
        "CF-IPCountry": "US",
    }
    request = Mock(headers=headers)
    device_data = await get_device_data(request)
    assert device_data["device"] == "Chrome on Windows"
    assert device_data["region"] == "Mountain View, US"

@pytest.mark.asyncio
async def test_is_username_valid():
    assert await is_username_valid("validuser") == {}
    assert "error" in await is_username_valid("sh")
    assert "error" in await is_username_valid("with space")
    assert "error" in await is_username_valid("invalid!")
    assert {} == await is_username_valid("axa_y")


@pytest.mark.asyncio
async def test_is_password_valid():
    assert await is_password_valid("ValidPass1") == {}
    assert "error" in await is_password_valid("short")
    assert "error" in await is_password_valid("nouppercase")
    assert "error" in await is_password_valid("NOLOWERCASE")
    assert "error" in await is_password_valid("")

def test_sanitize_username():
    assert sanitize_username(" User With Spaces ") == "Userwithspaces"
    assert sanitize_username("user-123_test") == "User123test"
    assert sanitize_username("!@#$%^&*()") == ""



# authtuna/core/exceptions.py (a new file)

class AuthTunaError(Exception):
    """Base exception for the AuthTuna library."""
    pass


class UserAlreadyExistsError(AuthTunaError):
    """Raised when trying to register a user that already exists."""
    pass


class InvalidCredentialsError(AuthTunaError):
    """Raised on login failure due to wrong username/password."""
    pass


class EmailNotVerifiedError(AuthTunaError):
    """Raised when a user tries to log in without a verified email."""
    pass


class UserNotFoundError(AuthTunaError):
    """Raised when a user is not found."""
    pass


class InvalidTokenError(AuthTunaError):
    """Raised when a token is invalid, used, or malformed."""
    pass


class TokenExpiredError(AuthTunaError):
    """
    Raised when a token has expired.
    Contains the new token if one was generated.
    """

    def __init__(self, message, new_token_id=None):
        super().__init__(message)
        self.new_token_id = new_token_id


class RateLimitError(AuthTunaError):
    """Raised when an action is performed too frequently."""
    pass


class SessionNotFoundError(AuthTunaError):
    """Raised when a session ID is not found in the database."""
    pass


class RoleNotFoundError(AuthTunaError):
    """Raised when a role is not found in the database."""
    pass


class PermissionNotFoundError(AuthTunaError):
    """Raised when a permission is not found in the database."""
    pass


class OperationForbiddenError(AuthTunaError):
    """Raised when an operation is forbidden."""
    pass



import pytest
from fastapi import Request, status
from fastapi.exceptions import HTTPException
from authtuna.integrations.fastapi_integration import get_current_user, get_user_ip, PermissionChecker
from unittest.mock import AsyncMock, MagicMock, patch

import types
import asyncio

@pytest.mark.asyncio
async def test_get_current_user_authenticated():
    user_obj = MagicMock(id='u1')
    request = MagicMock()
    request.state = types.SimpleNamespace(user_id='u1')
    with patch('authtuna.integrations.fastapi_integration.auth_service.users.get_by_id', new=AsyncMock(return_value=user_obj)):
        user = await get_current_user(request)
        test_getting_cached = await get_current_user(request)
        assert user.id == 'u1'
        assert user is user_obj
        assert request.state.user_object is user_obj
        assert test_getting_cached is user_obj

@pytest.mark.asyncio
async def test_get_current_user_unauthenticated():
    request = MagicMock()
    request.state = types.SimpleNamespace()
    with pytest.raises(HTTPException) as exc:
        await get_current_user(request)
    assert exc.value.status_code == status.HTTP_401_UNAUTHORIZED

@pytest.mark.asyncio
async def test_get_current_user_not_found():
    request = MagicMock()
    request.state = types.SimpleNamespace(user_id='u2')
    with patch('authtuna.integrations.fastapi_integration.auth_service.users.get_by_id', new=AsyncMock(return_value=None)):
        with pytest.raises(HTTPException) as exc:
            await get_current_user(request)
        print(exc.value.detail)
        assert exc.value.status_code == status.HTTP_401_UNAUTHORIZED

@pytest.mark.asyncio
async def test_get_current_user_error():
    request = MagicMock()
    request.state = types.SimpleNamespace(user_id='u3')
    with patch('authtuna.integrations.fastapi_integration.auth_service.users.get_by_id', new=AsyncMock(side_effect=Exception('fail'))):
        with pytest.raises(HTTPException) as exc:
            await get_current_user(request)
        assert exc.value.status_code == status.HTTP_500_INTERNAL_SERVER_ERROR


def test_get_user_ip():
    request = MagicMock()
    request.state = types.SimpleNamespace(user_ip_address='1.2.3.4')
    assert get_user_ip(request) == '1.2.3.4'

@pytest.mark.asyncio
async def test_permission_checker_and_mode():
    user = MagicMock(id='u1')
    request = MagicMock()
    request.state = types.SimpleNamespace(user_id='u1')
    # AND mode, missing permission
    checker = PermissionChecker('perm1', 'perm2', mode='AND')
    with patch('authtuna.integrations.fastapi_integration.auth_service.roles.has_permission', new=AsyncMock(return_value=False)):
        with pytest.raises(HTTPException) as exc:
            await checker(request, user)
        assert exc.value.status_code == status.HTTP_403_FORBIDDEN
    # OR mode, at least one permission
    checker = PermissionChecker('perm1', 'perm2', mode='OR')
    with patch('authtuna.integrations.fastapi_integration.auth_service.roles.has_permission', new=AsyncMock(side_effect=[False, True])):
        result = await checker(request, user)
        assert result is user
    # Scope from path param missing
    checker = PermissionChecker('perm', mode='AND', scope_from_path='org_id')
    request.path_params = {}
    with pytest.raises(HTTPException) as exc:
        await checker(request, user)
    assert exc.value.status_code == 500




import time

import pytest
from authtuna import settings
from authtuna.core.encryption import EncryptionUtils

@pytest.fixture
def encryption_utils():
    """Fixture to create an EncryptionUtils instance."""
    return EncryptionUtils()

def test_hash_and_verify_password(encryption_utils):
    """Test that password hashing and verification works correctly."""
    password = "test_password"
    hashed_password = encryption_utils.hash_password(password)
    assert encryption_utils.verify_password(password, hashed_password)
    assert not encryption_utils.verify_password("wrong_password", hashed_password)

def test_encrypt_and_decrypt_data(encryption_utils):
    """Test that data encryption and decryption works correctly."""
    data = b"test_data"
    encrypted_data = encryption_utils.encrypt_data(data)
    decrypted_data = encryption_utils.decrypt_data(encrypted_data.encode("utf-8"))
    assert decrypted_data == data.decode("utf-8")

def test_create_and_decode_jwt_token(encryption_utils):
    """Test that JWT token creation and decoding works correctly."""
    data = {"user_id": "123"}
    token = encryption_utils.create_jwt_token(data)
    token2 = encryption_utils.create_jwt_token(data, expires_delta=1)
    decoded_data = encryption_utils.decode_jwt_token(token)
    assert "user_id" in decoded_data
    assert decoded_data["user_id"] == data["user_id"]
    time.sleep(2)
    is_none = encryption_utils.decode_jwt_token(token2)
    assert is_none is None

def test_invalid_decode_jwt_token(encryption_utils):
    """Test that decoding an invalid JWT token raises an error."""
    invalid = "some_invalid_tkn"
    assert encryption_utils.decode_jwt_token(invalid) is None


def test_no_fernet_key():
    """Temporarily delete fernet keys to test error handling."""
    original_keys = settings.FERNET_KEYS
    settings.FERNET_KEYS = []
    with pytest.raises(ValueError):
        EncryptionUtils().encrypt_data(b"test")
    with pytest.raises(ValueError):
        EncryptionUtils().decrypt_data(b"test")
    settings.fernet_keys = original_keys



import pytest
from authtuna.core.database import CaseInsensitiveText, JsonType, User, Role, Permission, Session, Token, Base
from sqlalchemy import create_engine, Column, Integer, String
from sqlalchemy.orm import sessionmaker
import time
import types


def test_json_type_sqlite():
    engine = create_engine('sqlite:///:memory:')
    class TestModel(Base):
        __tablename__ = 'test_json'
        id = Column(Integer, primary_key=True)
        data = Column(JsonType)
    Base.metadata.create_all(engine)
    SessionLocal = sessionmaker(bind=engine)
    db = SessionLocal()
    db.add(TestModel(data={"foo": "bar"}))
    db.commit()
    result = db.query(TestModel).first()
    assert result.data == {"foo": "bar"}

# --- ORM Model Methods (sync logic only, async logic is tested in other files) ---
def test_user_repr_and_flags():
    user = User(id="1", username="testuser", email="test@example.com", is_active=True, email_verified=True, mfa_enabled=False)
    assert user.is_email_verified() is True
    assert user.has_role("admin") is False
    assert user.has_permission("perm") is False
    assert "testuser" in repr(user)

def test_role_and_permission_repr():
    role = Role(id=1, name="admin", description="desc")
    perm = Permission(id=1, name="perm", description="desc")
    assert "admin" in repr(role)
    assert "perm" in repr(perm)

def test_session_expiry_and_repr():
    session = Session(session_id="abc", user_id="1", region="r", device="d", active=True, ctime=time.time(), mtime=time.time(), etime=time.time()+100, e_abs_time=time.time()+100, create_ip="ip", last_ip="ip", random_string="rand", previous_random_strings=[])
    assert session.is_expired() is False
    assert isinstance(session.get_cookie_string(), str)

def test_token_validity_and_repr():
    token = Token(id="tok", purpose="test", user_id="1", ctime=time.time(), etime=time.time()+100, used=False)
    assert token.is_valid() is True
    assert "tok" == token.id

import pytest

@pytest.mark.asyncio
async def test_permission_manager_crud(auth_tuna_async):
    perm, created = await auth_tuna_async.permissions.get_or_create("perm:test", defaults={"description": "desc"})
    assert created is True
    perm2, created2 = await auth_tuna_async.permissions.get_or_create("perm:test")
    assert created2 is False
    assert perm.id == perm2.id
    found = await auth_tuna_async.permissions.get_by_name("perm:test")
    assert found.id == perm.id
    with pytest.raises(ValueError):
        await auth_tuna_async.permissions.create("perm:test")

@pytest.mark.asyncio
async def test_session_manager_crud(auth_tuna_async):
    user = await auth_tuna_async.users.create(email="sess@example.com", username="sessuser", password="pw", ip_address="1.1.1.1")
    session = await auth_tuna_async.sessions.create(user.id, "1.1.1.1", "region", "device")
    found = await auth_tuna_async.sessions.get_by_id(session.session_id)
    assert found.session_id == session.session_id
    all_sessions = await auth_tuna_async.sessions.get_all_for_user(user.id, session.session_id)
    assert any(s.session_id == session.session_id for s in all_sessions)
    await auth_tuna_async.sessions.terminate(session.session_id, "1.1.1.1")
    # Terminate non-existent session (should return False)
    assert not await auth_tuna_async.sessions.terminate("notfound", "1.1.1.1")
    # Terminate all for user (should not error)
    await auth_tuna_async.sessions.terminate_all_for_user(user.id, "1.1.1.1")

@pytest.mark.asyncio
async def test_token_manager_create(auth_tuna_async):
    user = await auth_tuna_async.users.create(email="tok@example.com", username="tokuser", password="pw", ip_address="1.1.1.1")
    token = await auth_tuna_async.tokens.create(user.id, "test-purpose", expiry_seconds=60)
    assert token.purpose == "test-purpose"
    assert token.is_valid()



"""
AuthTuna
========

A high-performance, async-first authorization and session management library for Python, with first-class FastAPI support.

AuthTuna provides a robust, multi-layered security foundation for modern web applications. It is designed for developers who need to build complex, multi-tenant systems without compromising on security or performance. The library combines a powerful hierarchical permission model with an advanced, stateful session management system to actively defend against a wide range of modern threats.

- FastAPI-first integration: ready-to-use dependencies and session middleware.
- Async SQLAlchemy models and managers for Users, Roles, Permissions, Sessions, Tokens, MFA, Social Accounts.
- Dual-state session model: server-side sessions + JWT cookie with rotating random_string and periodic DB verification.
- Session hijack detection: region/device fingerprint checks, IP tracking, automatic invalidation.
- Extensible RBAC with scoped permissions (e.g., "project:read" with scope_from_path).
- SQL-first design with PostgreSQL and SQLite support.

While the core is framework-agnostic and future adapters are possible, the officially supported and actively maintained integration is FastAPI.
"""

__version__ = "0.0.1"
__author__ = "shashstormer"
__description__ = "A robust, multi-layered security foundation for modern web applications"

import importlib
import threading

from fastapi import FastAPI

from .core.config import settings, init_settings


def _start_rpc_server_bg():
    if not settings.RPC_ENABLED:
        return
    import asyncio
    try:
        server_mod = importlib.import_module('authtuna.rpc.server')
        serve = getattr(server_mod, 'serve', None)
        if serve:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop.run_until_complete(serve())
    except Exception as e:
        import warnings
        warnings.warn(f"Failed to auto-start RPC server: {e}")


if getattr(settings, 'RPC_AUTOSTART', False):
    t = threading.Thread(target=_start_rpc_server_bg, daemon=True)
    t.start()


def init_app(app: FastAPI):
    """
    New wrapper to initialize AuthTuna with a FastAPI app instance and all routers.
    :param app:
    :return:
    """
    from authtuna.routers import admin_router, auth_router, social_router, mfa_router, ui_router
    from authtuna.middlewares import DatabaseSessionMiddleware
    from starlette.middleware.sessions import SessionMiddleware

    app.add_middleware(DatabaseSessionMiddleware)
    app.add_middleware(SessionMiddleware, settings.ENCRYPTION_PRIMARY_KEY.get_secret_value())
    app.include_router(auth_router)
    app.include_router(social_router)
    app.include_router(mfa_router)
    app.include_router(admin_router)
    app.include_router(ui_router)


__all__ = [
    "settings",
    "init_settings",
    "init_app"
]



import pytest
from fastapi import status
from httpx import AsyncClient
from authtuna.core.exceptions import InvalidCredentialsError

@pytest.mark.asyncio
async def test_signup_and_login(auth_tuna_async):
    """Test user signup and subsequent login."""
    # Signup a new user
    user, _ = await auth_tuna_async.signup(
        username="test_auth_user",
        email="test_auth@example.com",
        password="ValidPassword123",
        ip_address="127.0.0.1"
    )
    assert user is not None
    assert user.email == "test_auth@example.com"

    # Login with correct credentials
    loggedInUser, session = await auth_tuna_async.login(
        username_or_email="test_auth_user",
        password="ValidPassword123",
        ip_address="127.0.0.1",
        region="Test Region",
        device="Test Device"
    )
    assert loggedInUser is not None
    assert session is not None
    assert loggedInUser.id == user.id

    # Attempt login with incorrect password
    with pytest.raises(InvalidCredentialsError):
        await auth_tuna_async.login(
            username_or_email="test_auth_user",
            password="WrongPassword",
            ip_address="127.0.0.1",
            region="Test Region",
            device="Test Device"
        )

@pytest.mark.asyncio
async def test_signup_and_login_endpoints(fastapi_client: AsyncClient):
    # Signup page (GET)
    resp = await fastapi_client.get("/auth/signup")
    assert resp.status_code == status.HTTP_200_OK
    # Signup (POST)
    resp = await fastapi_client.post(
        "/auth/signup",
        json={"username": "testuser1", "email": "testuser1@example.com", "password": "pw123456"},
    )
    assert resp.status_code in (status.HTTP_201_CREATED, status.HTTP_202_ACCEPTED)
    # Duplicate signup
    resp2 = await fastapi_client.post(
        "/auth/signup",
        json={"username": "testuser1", "email": "testuser1@example.com", "password": "pw123456"},
    )
    assert resp2.status_code == status.HTTP_409_CONFLICT
    # Login page (GET)
    resp = await fastapi_client.get("/auth/login")
    assert resp.status_code == status.HTTP_200_OK
    # Login (POST)
    resp = await fastapi_client.post(
        "/auth/login",
        json={"username_or_email": "testuser1", "password": "pw123456"},
    )
    assert resp.status_code == status.HTTP_200_OK
    # Login with wrong password
    resp = await fastapi_client.post(
        "/auth/login",
        json={"username_or_email": "testuser1", "password": "wrongpw"},
    )
    assert resp.status_code == status.HTTP_401_UNAUTHORIZED

@pytest.mark.asyncio
async def test_logout_and_user_info(fastapi_client: AsyncClient):
    # Signup and login
    await fastapi_client.post(
        "/auth/signup",
        json={"username": "testuser2", "email": "testuser2@example.com", "password": "pw123456"},
    )
    resp = await fastapi_client.post(
        "/auth/login",
        json={"username_or_email": "testuser2", "password": "pw123456"},
    )
    token = resp.cookies.get("session_token")
    # User info (GET)
    resp = await fastapi_client.get("/auth/user-info", cookies={"session_token": token})
    assert resp.status_code == status.HTTP_200_OK
    # Logout (POST)
    resp = await fastapi_client.post("/auth/logout", cookies={"session_token": token})
    assert resp.status_code == status.HTTP_200_OK

@pytest.mark.asyncio
async def test_forgot_and_reset_password(fastapi_client: AsyncClient, monkeypatch):
    # Patch email sending to no-op
    monkeypatch.setattr("authtuna.helpers.mail.email_manager.send_password_reset_email", lambda *a, **k: None)
    monkeypatch.setattr("authtuna.helpers.mail.email_manager.send_password_change_email", lambda *a, **k: None)
    # Signup
    await fastapi_client.post(
        "/auth/signup",
        json={"username": "testuser3", "email": "testuser3@example.com", "password": "pw123456"},
    )
    # Forgot password (POST)
    resp = await fastapi_client.post(
        "/auth/forgot-password",
        json={"email": "testuser3@example.com"},
    )
    assert resp.status_code in (status.HTTP_202_ACCEPTED, status.HTTP_501_NOT_IMPLEMENTED)
    # Reset password (POST, invalid token)
    resp = await fastapi_client.post(
        "/auth/reset-password",
        json={"token": "badtoken", "new_password": "newpw123"},
    )
    assert resp.status_code in (status.HTTP_400_BAD_REQUEST, status.HTTP_501_NOT_IMPLEMENTED)

@pytest.mark.asyncio
async def test_verify_email_and_reset_page(fastapi_client: AsyncClient):
    # These endpoints render HTML, just check for 200 or error
    resp = await fastapi_client.get("/auth/verify?token=badtoken")
    assert resp.status_code == status.HTTP_200_OK
    resp = await fastapi_client.get("/auth/reset-password?token=badtoken")
    assert resp.status_code == status.HTTP_200_OK



import pytest
import time
from sqlalchemy import select

from authtuna.core.database import Session as DBSession, Token, SocialAccount
from authtuna.core.database import db_manager
from authtuna.core.encryption import encryption_utils


@pytest.mark.asyncio
async def test_session_is_valid_and_mismatches(auth_tuna_async):
    # Create a user and a session row directly
    user = await auth_tuna_async.users.create(
        email='sessv@example.com', username='sessv', password='ValidPassword123', ip_address='127.0.0.1'
    )
    async with db_manager.get_db() as db:
        sess = DBSession(
            session_id='s1', user_id=user.id, region='R', device='D', active=True,
            ctime=time.time(), mtime=time.time(), etime=time.time()+3600,
            e_abs_time=time.time()+86400, create_ip='127.0.0.1', last_ip='127.0.0.1',
            random_string=encryption_utils.gen_random_string(), previous_random_strings=[]
        )
        db.add(sess)
        await db.commit()
        # Valid match
        ok = await sess.is_valid(region='R', device='D', random_string=sess.random_string, db=db)
        assert ok is True
        # Region mismatch -> invalidated
        ok2 = await sess.is_valid(region='R2', device='D', random_string=sess.random_string, db=db)
        assert ok2 is False
        # Device mismatch -> invalidated
        sess.active = True
        ok3 = await sess.is_valid(region='R', device='D2', random_string=sess.random_string, db=db)
        assert ok3 is False
        # Random string mismatch -> invalidated
        sess.active = True
        ok4 = await sess.is_valid(region='R', device='D', random_string='bad', db=db)
        assert ok4 is False


@pytest.mark.asyncio
async def test_session_update_random_string_and_cookie(auth_tuna_async):
    user = await auth_tuna_async.users.create(
        email='sessrng@example.com', username='sessrng', password='ValidPassword123', ip_address='127.0.0.1'
    )
    async with db_manager.get_db() as db:
        sess = DBSession(
            session_id='s2', user_id=user.id, region='R', device='D', active=True,
            ctime=time.time(), mtime=time.time(), etime=time.time()+3600,
            e_abs_time=time.time()+86400, create_ip='127.0.0.1', last_ip='127.0.0.1',
            random_string=encryption_utils.gen_random_string(), previous_random_strings=[]
        )
        db.add(sess)
        await db.commit()
        first = sess.random_string
        new_rs = await sess.update_random_string()
        assert isinstance(new_rs, str) and new_rs != first
        # Cookie should be a JWT containing ids
        cookie = sess.get_cookie_string()
        payload = encryption_utils.decode_jwt_token(cookie)
        assert payload['session'] == sess.session_id
        assert payload['user_id'] == sess.user_id


@pytest.mark.asyncio
async def test_session_update_last_ip_and_terminate(auth_tuna_async):
    user = await auth_tuna_async.users.create(
        email='sessip@example.com', username='sessip', password='ValidPassword123', ip_address='127.0.0.1'
    )
    async with db_manager.get_db() as db:
        sess = DBSession(
            session_id='s3', user_id=user.id, region='R', device='D', active=True,
            ctime=time.time(), mtime=time.time(), etime=time.time()+3600,
            e_abs_time=time.time()+86400, create_ip='127.0.0.1', last_ip='127.0.0.1',
            random_string=encryption_utils.gen_random_string(), previous_random_strings=[]
        )
        db.add(sess)
        await db.commit()
        await sess.update_last_ip('127.0.0.2', db=db)
        assert sess.last_ip == '127.0.0.2'
        await sess.terminate('127.0.0.2', db=db)
        assert sess.active is False


@pytest.mark.asyncio
async def test_token_is_valid_and_mark_used(auth_tuna_async):
    suffix = str(int(time.time() * 1000000))
    user = await auth_tuna_async.users.create(
        email=f'tok{suffix}@example.com', username=f'tok{suffix}', password='ValidPassword123', ip_address='127.0.0.1'
    )
    token = await auth_tuna_async.tokens.create(user.id, 'email_verification')
    assert token.is_valid() is True
    async with db_manager.get_db() as db:
        await token.mark_used('127.0.0.1', db=db)
        assert token.used is True


@pytest.mark.asyncio
async def test_social_account_model_minimal(auth_tuna_async):
    suffix = str(int(time.time() * 1000000))
    user = await auth_tuna_async.users.create(
        email=f'sacc{suffix}@example.com', username=f'sacc{suffix}', password='ValidPassword123', ip_address='127.0.0.1'
    )
    async with db_manager.get_db() as db:
        sacc = SocialAccount(
            user_id=user.id, provider='google', provider_user_id='gid', token_type='bearer', access_token='at'
        )
        db.add(sacc)
        await db.commit()
        # Fetch back to ensure persistence
        row = (await db.execute(select(SocialAccount))).scalars().first()
        assert row is not None and row.provider == 'google'



import pytest
from authtuna.core.social import get_social_provider, oauth
from fastapi import status
from httpx import AsyncClient
from unittest.mock import patch, AsyncMock

def test_get_social_provider_valid():
    # These providers may or may not be registered depending on settings, but the function should not error
    for provider in ["google", "github"]:
        client = get_social_provider(provider)
        assert client is None or hasattr(client, 'authorize_redirect')

def test_get_social_provider_invalid():
    assert get_social_provider("notarealprovider") is None

def test_oauth_registry_is_accessible():
    # The oauth object should be present and have a registry dict
    print(oauth)
    assert hasattr(oauth, '_registry')
    assert isinstance(oauth._registry, dict)

@pytest.mark.asyncio
async def test_social_login_provider_not_found(fastapi_client: AsyncClient):
    resp = await fastapi_client.get("/auth/notarealprovider/login")
    assert resp.status_code == status.HTTP_404_NOT_FOUND
    resp = await fastapi_client.get("/auth/notarealprovider/callback")
    assert resp.status_code == status.HTTP_404_NOT_FOUND



from typing import Optional, List

from fastapi import APIRouter, Depends, HTTPException, status, Query, Request
from fastapi.responses import HTMLResponse
from pydantic import BaseModel, Field
from starlette.templating import Jinja2Templates

from authtuna import settings
from authtuna.core.database import User
from authtuna.core.exceptions import RoleNotFoundError, PermissionNotFoundError, UserNotFoundError, \
    OperationForbiddenError
from authtuna.integrations.fastapi_integration import get_current_user, auth_service, PermissionChecker
from authtuna.routers.auth import RoleInfo

# --- Router Setup ---
# A router-level dependency ensures that users have basic admin panel access.
# Specific, more powerful permissions are checked on each endpoint individually.
router = APIRouter(
    prefix="/admin", tags=["Administration"], dependencies=[Depends(PermissionChecker("admin:access:panel"))],
)

templates = Jinja2Templates(directory=settings.HTML_TEMPLATE_DIR)


# --- Pydantic Models (Unchanged) ---

class RoleCreate(BaseModel):
    name: str = Field(..., description="The unique name for the new role.")
    description: Optional[str] = Field(None, description="A brief description of the role's purpose.")
    level: Optional[int] = Field(None, description="An optional integer hierarchy level for the role.")


class PermissionCreate(BaseModel):
    name: str = Field(..., description="The unique name for the new permission (e.g., 'timeline:edit:content').")
    description: Optional[str] = Field(None, description="A brief description of what the permission allows.")


class AssignPermissionToRole(BaseModel):
    permission_name: str


class AssignRoleToUser(BaseModel):
    user_id: str
    role_name: str
    scope: str = Field("global", description="The scope in which the role is granted (e.g., 'global', 'timeline:123').")


class UserSearchResult(BaseModel):
    id: str
    username: str
    email: str
    is_active: bool
    mfa_enabled: bool

    class Config:
        from_attributes = True


class UserSuspend(BaseModel):
    reason: str = Field("No reason provided.", description="The reason for suspending the user.")


class AuditEventResponse(BaseModel):
    event_type: str
    timestamp: float
    ip_address: Optional[str]
    details: Optional[dict]

    class Config:
        from_attributes = True


class GrantRoleAssignPermission(BaseModel):
    assigner_role_name: str
    assignable_role_name: str


class GrantPermissionGrantPermission(BaseModel):
    granter_role_name: str
    grantable_permission_name: str

class PermissionInfo(BaseModel):
    name: str

    class Config:
        from_attributes = True


class UserInRole(BaseModel):
    username: str
    scope: str


class RoleDetail(RoleInfo):
    permissions: List[PermissionInfo]
    users: List[UserInRole]

@router.get(
    "/users/search", response_model=List[UserSearchResult], summary="Search and Filter Users",
    dependencies=[Depends(PermissionChecker("admin:manage:users"))]  # ADDED: Specific permission
)
async def search_users_endpoint(
        identity: Optional[str] = Query(None, description="Search by email or username."),
        role: Optional[str] = Query(None, description="Filter by a role the user has."),
        scope: Optional[str] = Query(None, description="Filter by a scope the user has a role in."),
        is_active: Optional[bool] = Query(None, description="Filter by user's active status."), skip: int = 0,
        limit: int = 50
):
    """Provides advanced, flexible filtering for users."""
    users = await auth_service.users.search_users(
        identity=identity, role=role, scope=scope, is_active=is_active, skip=skip, limit=limit
    )
    return users

@router.get(
    "/users/{user_id}", response_class=HTMLResponse, summary="Serve the User Detail Page"
)
async def serve_user_detail_page(request: Request):
    """Serves the HTML page for viewing a single user's details."""
    return templates.TemplateResponse("admin_user_detail.html", {"request": request})

@router.get("/assignable-roles", dependencies=[Depends(PermissionChecker("admin:manage:roles"))])
async def get_my_assignable_roles(current_user: User = Depends(get_current_user), ):
    """
    Returns a list of roles the current user can assign to themselves.
    """
    assignable_roles = await auth_service.roles.get_self_assignable_roles(current_user.id)
    return assignable_roles

@router.get(
    "/users/{user_id}/assignable-roles", summary="Get Roles Assignable by Current Admin",
    dependencies=[Depends(PermissionChecker("admin:manage:roles"))]
)
async def get_assignable_roles_for_user(user_id: str, admin_user: User = Depends(get_current_user)):
    """
    Retrieves a list of roles that the currently authenticated admin
    is permitted to assign to the specified user.
    """
    try:
        assignable_roles = await auth_service.roles.get_assignable_roles_for_user(
            target_user_id=user_id, assigning_user=admin_user
        )
        return assignable_roles
    except UserNotFoundError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))


@router.get(
    "/users/{user_id}/details-data", summary="Get Detailed User Information",
    dependencies=[Depends(PermissionChecker("admin:manage:users"))]
)
async def get_user_details_data(user_id: str):
    """Retrieves detailed information about a single user, including their roles."""
    user = await auth_service.users.get_by_id(user_id, with_relations=True)
    if not user:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="User not found.")

    roles = await auth_service.roles.get_user_roles_with_scope(user.id)
    can_assign = await auth_service.roles.get_self_assignable_roles(user.id)
    return {
        "id": user.id,
        "email": user.email,
        "is_active": user.is_active,
        "mfa_enabled": user.mfa_enabled,
        "username": user.username,
        "roles": roles,
        "can_assign": can_assign,
    }


@router.get(
    "/roles/{role_name}", response_class=HTMLResponse, summary="Serve the Role Detail Page"
)
async def serve_role_detail_page(request: Request):
    """Serves the HTML page for viewing a single role's details."""
    return templates.TemplateResponse("admin_role_detail.html", {"request": request})


@router.get(
    "/roles/{role_name}/details-data", summary="Get Detailed Role Information",
    dependencies=[Depends(PermissionChecker("admin:manage:roles"))]
)
async def get_role_details_data(role_name: str):
    """Retrieves detailed information about a single role, including permissions and users."""
    role = await auth_service.roles.get_by_name(role_name)
    if not role:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Role not found.")

    users = await auth_service.roles.get_users_for_role(role_name)

    return {
        "name": role.name, "description": role.description, "level": role.level, "permissions": role.permissions,
        "users": users
    }


@router.get(
    "/dashboard", response_class=HTMLResponse, summary="Serve the Admin Dashboard HTML Page"
)
async def serve_admin_dashboard(request: Request):
    """Serves the main admin dashboard HTML page."""
    return templates.TemplateResponse("admin_dashboard.html", {"request": request})


@router.get(
    "/roles", summary="List All Roles", dependencies=[Depends(PermissionChecker("admin:manage:roles"))]
)
async def list_roles():
    """Retrieves a list of all roles in the system."""
    roles = await auth_service.roles.get_all_roles()
    return roles


# --- Admin Endpoints ---


@router.post(
    "/users/{user_id}/suspend", response_model=UserSearchResult, summary="Suspend a User Account",
    dependencies=[Depends(PermissionChecker("admin:manage:users"))]  # ADDED: Specific permission
)
async def suspend_user(user_id: str, payload: UserSuspend, admin_user: User = Depends(get_current_user)):
    """Suspends a user, preventing them from logging in."""
    try:
        user = await auth_service.users.suspend_user(user_id, admin_id=admin_user.id, reason=payload.reason)
        return user
    except UserNotFoundError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))


@router.post(
    "/users/{user_id}/unsuspend", response_model=UserSearchResult, summary="Unsuspend a User Account",
    dependencies=[Depends(PermissionChecker("admin:manage:users"))]  # ADDED: Specific permission
)
async def unsuspend_user(user_id: str, payload: UserSuspend, admin_user: User = Depends(get_current_user)):
    """Reactivates a previously suspended user."""
    try:
        user = await auth_service.users.unsuspend_user(user_id, admin_id=admin_user.id, reason=payload.reason)
        return user
    except UserNotFoundError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))


@router.get(
    "/users/{user_id}/audit-log", response_model=List[AuditEventResponse], summary="Get User Audit Log",
    dependencies=[Depends(PermissionChecker("admin:manage:users"))]  # ADDED: Specific permission
)
async def get_user_audit_log(user_id: str, skip: int = 0, limit: int = 25):
    """Retrieves the security audit trail for a specific user."""
    events = await auth_service.audit.get_events_for_user(user_id, skip=skip, limit=limit)
    return events


@router.post(
    "/roles", status_code=status.HTTP_201_CREATED, dependencies=[Depends(PermissionChecker("admin:manage:roles"))]
    # ADDED: Specific permission
)
async def create_role(role_data: RoleCreate):
    """Creates a new role in the system with an optional level."""
    try:
        await auth_service.roles.create(
            name=role_data.name, description=role_data.description, level=role_data.level
        )
        return {"message": f"Role '{role_data.name}' created successfully."}
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))


@router.post(
    "/permissions", status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(PermissionChecker("admin:manage:permissions"))]  # ADDED: Specific permission
)
async def create_permission(permission_data: PermissionCreate):
    """Creates a new permission in the system."""
    try:
        await auth_service.permissions.get_or_create(permission_data.name,
                                                     defaults={"description": permission_data.description})
        return {"message": f"Permission '{permission_data.name}' created successfully."}
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))


@router.post(
    "/roles/{role_name}/permissions", dependencies=[Depends(PermissionChecker("admin:manage:roles"))]
    # ADDED: Specific permission
)
async def add_permission_to_role(role_name: str, payload: AssignPermissionToRole,
                                 admin_user: User = Depends(get_current_user)):
    """Assigns an existing permission to an existing role."""
    try:
        await auth_service.roles.add_permission_to_role(role_name, payload.permission_name, admin_user.id)
        return {"message": f"Permission '{payload.permission_name}' added to role '{role_name}'."}
    except (RoleNotFoundError, PermissionNotFoundError) as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))


@router.post(
    "/users/roles/assign", dependencies=[Depends(PermissionChecker("admin:manage:roles"))]  # ADDED: Specific permission
)
async def assign_role_to_user(payload: AssignRoleToUser, admin_user: User = Depends(get_current_user)):
    """Assigns a role to a user."""
    try:
        await auth_service.roles.assign_to_user(
            user_id=payload.user_id, role_name=payload.role_name, scope=payload.scope, assigner_id=admin_user.id
        )
        return {"message": f"Role '{payload.role_name}' assigned to user {payload.user_id} in scope '{payload.scope}'."}
    except (UserNotFoundError, RoleNotFoundError) as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))
    except OperationForbiddenError as e:
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail=str(e))


@router.post(
    "/users/roles/revoke", dependencies=[Depends(PermissionChecker("admin:manage:roles"))]  # ADDED: Specific permission
)
async def revoke_role_from_user(payload: AssignRoleToUser, admin_user: User = Depends(get_current_user)):
    """Revokes a role from a user within a specific scope."""
    try:
        success = await auth_service.roles.revoke_user_role_by_scope(
            user_id=payload.user_id, role_name=payload.role_name, scope=payload.scope, revoker_id=admin_user.id, )
        if not success:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Role assignment not found.")
        return {"message": f"Role '{payload.role_name}' revoked from user {payload.user_id}."}
    except (RoleNotFoundError, UserNotFoundError) as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))
    except OperationForbiddenError as e:
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail=str(e))


@router.delete(
    "/roles/{role_name}", dependencies=[Depends(PermissionChecker("admin:manage:roles"))]  # ADDED: Specific permission
)
async def delete_role(role_name: str, admin_user: User = Depends(get_current_user)):
    """Deletes a role from the system."""
    try:
        await auth_service.roles.delete_role(role_name=role_name, deleter_id=admin_user.id)
        return {"message": f"Role '{role_name}' has been deleted."}
    except RoleNotFoundError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))
    except OperationForbiddenError as e:
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail=str(e))


@router.post(
    "/roles/grants/assign-role", status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(PermissionChecker("admin:manage:roles"))]  # ADDED: Specific permission
)
async def grant_role_assignment_permission(payload: GrantRoleAssignPermission):
    """Authorizes a role to be able to assign another role."""
    try:
        granter_role, assignable_role = await auth_service.roles.grant_relationship(
            granter_role_name=payload.assigner_role_name, grantable_name=payload.assignable_role_name,
            grantable_manager=auth_service.roles, relationship_attr="can_assign_roles"
        )
        return {"message": f"Role '{granter_role.name}' can now assign role '{assignable_role.name}'."}
    except (RoleNotFoundError, PermissionNotFoundError) as e:
        raise HTTPException(status_code=404, detail=str(e))


@router.post(
    "/roles/grants/grant-permission", status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(PermissionChecker("admin:manage:permissions"))]  # ADDED: Specific permission
)
async def grant_permission_granting_permission(payload: GrantPermissionGrantPermission):
    """Authorizes a role to grant a specific permission to other roles."""
    try:
        granter_role, permission = await auth_service.roles.grant_relationship(
            granter_role_name=payload.granter_role_name, grantable_name=payload.grantable_permission_name,
            grantable_manager=auth_service.permissions, relationship_attr="can_grant_permissions"
        )
        return {"message": f"Role '{granter_role.name}' can now grant permission '{permission.name}'."}
    except (RoleNotFoundError, PermissionNotFoundError) as e:
        raise HTTPException(status_code=404, detail=str(e))



import pytest
from unittest.mock import patch, AsyncMock, Mock, PropertyMock
from fastapi import status
from httpx import AsyncClient
from authlib.integrations.starlette_client import OAuthError


@pytest.mark.asyncio
async def test_social_login_redirect_google(fastapi_client: AsyncClient):
    async def fake_authorize_redirect(request, redirect_uri):
        class FakeResp:
            status_code = status.HTTP_302_FOUND
        return FakeResp()

    fake_provider = Mock()
    fake_provider.authorize_redirect = AsyncMock(side_effect=fake_authorize_redirect)

    with patch('authtuna.routers.social.get_social_provider', return_value=fake_provider):
        resp = await fastapi_client.get('/auth/google/login')
        assert resp.status_code in (status.HTTP_302_FOUND, status.HTTP_200_OK)


@pytest.mark.asyncio
async def test_social_callback_new_user_google_success(fastapi_client: AsyncClient):
    # Mock provider that returns token and userinfo without touching request.session
    fake_provider = Mock()
    fake_provider.authorize_access_token = AsyncMock(return_value={'access_token': 'tok', 'token_type': 'bearer'})
    fake_provider.userinfo = AsyncMock(return_value={'sub': 'google_user_1', 'email': 'newg@example.com', 'name': 'New G'})

    with patch('authtuna.routers.social.get_social_provider', return_value=fake_provider):
        # Also avoid sending emails in test
        with patch('authtuna.routers.social.email_manager') as mock_email_mgr:
            mock_email_mgr.send_new_social_account_connected_email = AsyncMock()
            mock_email_mgr.send_welcome_email = AsyncMock()
            resp = await fastapi_client.get('/auth/google/callback')
            assert resp.status_code in (status.HTTP_302_FOUND, status.HTTP_307_TEMPORARY_REDIRECT)


@pytest.mark.asyncio
async def test_social_callback_oauth_error(fastapi_client: AsyncClient):
    fake_provider = Mock()
    fake_provider.authorize_access_token = AsyncMock(side_effect=OAuthError(description='bad'))

    with patch('authtuna.routers.social.get_social_provider', return_value=fake_provider):
        resp = await fastapi_client.get('/auth/google/callback')
        assert resp.status_code == status.HTTP_400_BAD_REQUEST


@pytest.mark.asyncio
async def test_social_login_unknown_provider(fastapi_client: AsyncClient):
    resp = await fastapi_client.get('/auth/unknownprovider/login')
    assert resp.status_code == 404
    assert "not found" in resp.text.lower()


@pytest.mark.asyncio
async def test_social_callback_unknown_provider(fastapi_client: AsyncClient):
    resp = await fastapi_client.get('/auth/unknownprovider/callback')
    assert resp.status_code == 404
    assert "not found" in resp.text.lower()


@pytest.mark.asyncio
async def test_social_callback_unsupported_provider(fastapi_client: AsyncClient):
    fake_provider = Mock()
    fake_provider.authorize_access_token = AsyncMock(return_value={'access_token': 'tok', 'token_type': 'bearer'})
    with patch('authtuna.routers.social.get_social_provider', return_value=fake_provider):
        resp = await fastapi_client.get('/auth/unknown/callback')
        assert resp.status_code == 400 or resp.status_code == 404

@pytest.mark.asyncio
async def test_social_callback_new_user_username_sanitization(fastapi_client: AsyncClient):
    fake_provider = Mock()
    fake_provider.authorize_access_token = AsyncMock(return_value={'access_token': 'tok', 'token_type': 'bearer'})
    fake_provider.userinfo = AsyncMock(return_value={'sub': 'google_user_2', 'email': 'new2@example.com', 'name': None})
    with patch('authtuna.routers.social.get_social_provider', return_value=fake_provider):
        with patch('authtuna.routers.social.email_manager') as mock_email_mgr:
            mock_email_mgr.send_new_social_account_connected_email = AsyncMock()
            mock_email_mgr.send_welcome_email = AsyncMock()
            resp = await fastapi_client.get('/auth/google/callback')
            assert resp.status_code in (status.HTTP_302_FOUND, status.HTTP_307_TEMPORARY_REDIRECT)



import base64
import json
import logging
import secrets
import string
import time
from typing import Optional, Sequence
import bcrypt
from cryptography.fernet import Fernet, MultiFernet
from jose import jwt, JWTError
from authtuna.core.config import settings

logger = logging.getLogger(__name__)


class EncryptionUtils:
    """
    A utility class for handling both password hashing and data encryption with key rotation.
    """

    def __init__(self):
        """
        Initializes the EncryptionUtils with a list of base64-encoded Fernet keys.
        The first key in the list is the primary key for encryption.
        The subsequent keys are used for decryption of older data.
        """
        fernet_keys = settings.FERNET_KEYS
        self.fernet_initialized = True
        if not fernet_keys:
            self.fernet_initialized = False
        if fernet_keys is None or not isinstance(fernet_keys, Sequence) or not fernet_keys:
            logger.debug("RANDOM GENERATED FERNET KEY: " + self.generate_new_key())
            raise ValueError("A sequence of at least one Fernet key must be provided.")
        self.fernet_keys = [Fernet(k.get_secret_value().encode('utf-8')) for k in fernet_keys]
        self.multi_fernet = MultiFernet(self.fernet_keys)
        self.jwt_secret = settings.JWT_SECRET_KEY.get_secret_value()
        self.jwt_algorithm = settings.ALGORITHM

    @staticmethod
    def hash_password(password: str, bcrypt_rounds: int = 12) -> str:
        """
        Hashes a password using bcrypt.
        """
        salt = bcrypt.gensalt(bcrypt_rounds)
        hashed_password = bcrypt.hashpw(password.encode('utf-8'), salt)
        return hashed_password.decode('utf-8')

    @staticmethod
    def verify_password(password: str, hashed_password: str) -> bool:
        """
        Verifies a plain-text password against a bcrypt hash.
        """
        return bcrypt.checkpw(password.encode('utf-8'), hashed_password.encode('utf-8'))

    def encrypt_data(self, data: bytes) -> str:
        """
        Encrypts data using the primary (newest) Fernet key.
        """
        if not self.fernet_initialized:
            raise ValueError("Fernet is not initialized.")
        return self.fernet_keys[0].encrypt(data).decode('utf-8')

    def decrypt_data(self, data: bytes) -> str:
        """
        Decrypts data by attempting all keys in the rotation.
        """
        if not self.fernet_initialized:
            raise ValueError("Fernet is not initialized.")
        return self.multi_fernet.decrypt(data).decode('utf-8')

    @staticmethod
    def gen_random_string(length: int = 8, symbol_set: Optional[Sequence] = None):
        """
        Generates a cryptographically secure random string.
        """
        symbol_set = (string.ascii_letters + string.digits) if symbol_set is None else symbol_set
        return ''.join(secrets.choice(symbol_set) for _ in range(length))

    @staticmethod
    def generate_new_key() -> str:
        """
        Generates a new, URL-safe base64-encoded Fernet key.
        """
        return base64.urlsafe_b64encode(secrets.token_bytes(32)).decode('utf-8')

    def create_jwt_token(self, data: dict, expires_delta: Optional[int] = None) -> str:
        """
        Creates a new JWT token with an optional expiration time.

        Args:
            data (dict): The payload to be encoded in the JWT.
            expires_delta (Optional[int]): The lifetime of the token in seconds.

        Returns:
            str: The signed JWT token string.
        """
        to_encode = data.copy()
        if expires_delta:
            expire = time.time() + expires_delta
        else:
            expire = time.time() + settings.SESSION_LIFETIME_SECONDS

        to_encode.update({"exp": expire})
        return jwt.encode({"session": self.encrypt_data(json.dumps(to_encode).encode("utf-8"))}, self.jwt_secret, algorithm=self.jwt_algorithm)

    def decode_jwt_token(self, token: str) -> Optional[dict]:
        """
        Decodes a JWT token and returns its payload if valid.

        Args:
            token (str): The JWT token string to decode.

        Returns:
            Optional[dict]: The decoded payload, or None if decoding fails.
        """
        try:
            payload = jwt.decode(token, self.jwt_secret, algorithms=[self.jwt_algorithm])
            payload = json.loads(self.decrypt_data(payload["session"]))
            if "exp" in payload and time.time() > payload["exp"]:
                return None
            return payload
        except JWTError as e:
            logger.debug(f"JWT decoding failed: {e}")
            return None


encryption_utils = EncryptionUtils()



from authtuna.routers.auth import router as auth_router
from authtuna.routers.social import router as social_router
from authtuna.routers.mfa import router as mfa_router
from authtuna.routers.admin import router as admin_router
from authtuna.routers.ui import router as ui_router



import time
from typing import Optional, Tuple, List, Dict, Any, Union

import pyotp
from sqlalchemy import or_, select, func, delete, desc
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import joinedload, selectinload

from authtuna.core import encryption_utils
from authtuna.core.config import settings
from authtuna.core.database import (
    DatabaseManager, User, Role, Permission, DeletedUser,
    Session as DBSession, Token, user_roles_association, role_permissions_association, Session, AuditEvent
)
from authtuna.core.exceptions import (
    UserAlreadyExistsError, InvalidCredentialsError, EmailNotVerifiedError,
    InvalidTokenError, TokenExpiredError, RateLimitError, UserNotFoundError,
    SessionNotFoundError, RoleNotFoundError, PermissionNotFoundError, OperationForbiddenError
)
from authtuna.core.mfa import MFAManager


class UserManager:
    """Manages all CRUD and business logic operations related to Users, asynchronously."""

    def __init__(self, db_manager: DatabaseManager):
        self._db_manager = db_manager

    async def get_by_id(self, user_id: str, with_relations: bool = True, db: AsyncSession = None) -> Optional[User]:
        async def _get(session: AsyncSession):
            stmt = select(User).where(User.id == user_id)
            if with_relations:
                stmt = stmt.options(
                    selectinload(User.roles).selectinload(Role.permissions),
                    selectinload(User.roles).selectinload(Role.can_assign_roles)
                )
            result = await session.execute(stmt)
            return result.unique().scalar_one_or_none()

        if db:
            return await _get(db)
        async with self._db_manager.get_db() as db:
            return await _get(db)

    async def get_by_email(self, email: str) -> Optional[User]:
        async with self._db_manager.get_db() as db:
            stmt = select(User).where(User.email == email)
            result = await db.execute(stmt)
            return result.unique().scalar_one_or_none()

    async def get_by_username(self, username: str) -> Optional[User]:
        async with self._db_manager.get_db() as db:
            stmt = select(User).where(User.username == username)
            result = await db.execute(stmt)
            return result.unique().scalar_one_or_none()

    async def list(self, skip: int = 0, limit: int = 100) -> List[User]:
        async with self._db_manager.get_db() as db:
            stmt = select(User).offset(skip).limit(limit)
            result = (await db.execute(stmt)).unique()
            return list(result.scalars().all())

    async def create(self, email: str, username: str, password: Optional[str] = None, ip_address: str = 'system',
                     **kwargs) -> User:
        """Creates a user, sets password, and logs audit event in a single atomic transaction."""
        async with self._db_manager.get_db() as db:
            stmt = select(User).where((User.email == email) | (User.username == username))
            existing_user = (await db.execute(stmt)).unique().scalar_one_or_none()
            if existing_user:
                raise UserAlreadyExistsError("A user with this email or username already exists.")

            new_user = User(
                id=encryption_utils.gen_random_string(32), email=email, username=username, **kwargs
            )
            db.add(new_user)

            if password:
                await new_user.set_password(password, ip_address, self._db_manager, db)

            await self._db_manager.log_audit_event(
                new_user.id, "USER_CREATED", ip_address,
                {"by": "system" if ip_address == 'system' else 'user_signup'}, db=db
            )

            await db.commit()
            await db.refresh(new_user)
            return new_user

    async def update(self, user_id: str, update_data: Dict[str, Any], ip_address: str = 'system') -> User:
        async with self._db_manager.get_db() as db:
            user = await db.get(User, user_id)
            if not user:
                raise UserNotFoundError("User not found.")

            for key, value in update_data.items():
                if hasattr(user, key) and key not in ['id', 'password_hash']:
                    setattr(user, key, value)

            await self._db_manager.log_audit_event(user_id, "USER_UPDATED", ip_address,
                                                   {"fields_changed": list(update_data.keys())}, db=db)
            await db.commit()
            await db.refresh(user)
            return user

    async def delete(self, user_id: str, ip_address: str = 'system') -> None:
        async with self._db_manager.get_db() as db:
            user = await db.get(User, user_id)
            if not user:
                raise UserNotFoundError("User not found.")

            user_data = {c.name: getattr(user, c.name) for c in user.__table__.columns}
            archived_user = DeletedUser(user_id=user.id, email=user.email, data=user_data)
            db.add(archived_user)

            await db.delete(user)
            await self._db_manager.log_audit_event("system", "USER_DELETED", ip_address, {"archived": True}, db=db)
            await db.commit()

    async def set_password(self, user_id: str, new_password: str, ip_address: str):
        async with self._db_manager.get_db() as db:
            user = await db.get(User, user_id)
            if not user:
                raise UserNotFoundError("User not found.")
            await user.set_password(new_password, ip_address, self._db_manager, db)
            await db.commit()

    async def suspend_user(self, user_id: str, admin_id: str, reason: str = "No reason provided.") -> User:
        """Suspends a user, preventing them from logging in."""
        async with self._db_manager.get_db() as db:
            user = await db.get(User, user_id)
            if not user: raise UserNotFoundError("User not found.")
            user.is_active = False
            await self._db_manager.log_audit_event(
                user_id, "USER_SUSPENDED", "system", {"by": admin_id, "reason": reason}, db=db
            )
            await db.commit()
            await db.refresh(user)
            return user

    async def unsuspend_user(self, user_id: str, admin_id: str, reason: str = "No reason provided.") -> User:
        """Reactivates a suspended user."""
        async with self._db_manager.get_db() as db:
            user = await db.get(User, user_id)
            if not user: raise UserNotFoundError("User not found.")
            user.is_active = True
            await self._db_manager.log_audit_event(
                user_id, "USER_UNSUSPENDED", "system", {"by": admin_id, "reason": reason}, db=db
            )
            await db.commit()
            await db.refresh(user)
            return user

    # function to search users with filtering by email or username or roles he has or scopes he has
    async def search_users(
            self,
            *,
            identity: Optional[str] = None,
            role: Optional[str] = None,
            scope: Optional[str] = None,
            is_active: Optional[bool] = None,
            skip: int = 0,
            limit: int = 100
    ) -> List[User]:
        """
        Provides advanced, flexible filtering for users based on specific criteria.
        All filter parameters are optional and are combined with AND logic.
        """
        async with self._db_manager.get_db() as db:
            stmt = select(User)
            filters = []

            if identity:
                filters.append(or_(
                    User.email == identity,
                    User.username.ilike(f"%{identity}%")
                ))
            if is_active is not None:
                filters.append(User.is_active == is_active)

            if role or scope:
                stmt = stmt.join(user_roles_association).join(Role)
                if role:
                    filters.append(Role.name.ilike(f"%{role}%"))
                if scope:
                    filters.append(user_roles_association.c.scope.ilike(f"%{scope}%"))

            if filters:
                stmt = stmt.where(*filters)

            stmt = stmt.offset(skip).limit(limit).distinct()
            result = await db.execute(stmt)
            return list(result.unique().scalars().all())

    async def basic_search_users(self, *, identity: Optional[str] = None, skip: int = 0, limit: int = 100) -> List[
        Dict[str, str]]:
        """
        Provides basic filtering for users based on username or email.
        This function readcts the email for privacy and gives only username and userid.
        This function can be used to implement a search user route in your codebase.
        """
        users = await self.search_users(identity=identity, skip=skip, limit=limit)
        basic_users = []
        if users:
            for user in users:
                basic_users.append({
                    "user_id": user.id,
                    "username": user.username,
                })
        return basic_users


class RoleManager:
    def __init__(self, db_manager: DatabaseManager):
        self._db_manager = db_manager

    async def get_all_roles(self) -> List[Role]:
        """Fetches all roles from the database."""
        async with self._db_manager.get_db() as db:
            stmt = select(Role).order_by(Role.level.desc().nullslast(), Role.name)
            result = await db.execute(stmt)
            return list(result.unique().scalars().all())

    async def get_assignable_roles_for_user(self, target_user_id: str, assigning_user: User) -> List[Role]:
        """
        Determines which roles an admin can assign, preventing privilege escalation.
        This function correctly checks all three authorization pathways:
        1. Direct permission override (e.g., 'roles:assign:SomeRole').
        2. Direct role-to-role assignment grants.
        3. Role level hierarchy (admin's level must be higher).
        """
        async with self._db_manager.get_db() as db:
            target_user = await db.get(User, target_user_id)
            if not target_user:
                raise UserNotFoundError("Target user not found.")
            all_roles = await self.get_all_roles()
            admin_max_level = -1
            if assigning_user.roles:
                admin_max_level = max(
                    (role.level for role in assigning_user.roles if role.level is not None),
                    default=-1
                )
            assignable_roles = []
            for role_to_check in all_roles:
                has_sufficient_level = False
                if role_to_check.level is not None and admin_max_level > role_to_check.level:
                    has_sufficient_level = True
                has_direct_grant = False
                for admin_role in assigning_user.roles:
                    if any(assignable.id == role_to_check.id for assignable in admin_role.can_assign_roles):
                        has_direct_grant = True
                        break
                required_permission = f"roles:assign:{role_to_check.name}"
                has_permission_override = await self.has_permission(assigning_user.id, required_permission, db=db)
                if has_sufficient_level or has_direct_grant or has_permission_override:
                    assignable_roles.append(role_to_check)
            return assignable_roles

    async def get_permission_scopes(self, user_id: str, permission_name: str, db: AsyncSession) -> List[str]:
        """
        Retrieves all scopes in which a user has a specific permission.
        """
        stmt = select(user_roles_association.c.scope).distinct().join_from(
            user_roles_association, Role, user_roles_association.c.role_id == Role.id
        ).join(
            role_permissions_association, Role.id == role_permissions_association.c.role_id
        ).join(
            Permission, role_permissions_association.c.permission_id == Permission.id
        ).where(
            user_roles_association.c.user_id == user_id,
            Permission.name == permission_name
        )
        result = await db.execute(stmt)
        return [row[0] for row in result.all()]

    async def get_self_assignable_roles(self, user_id: str) -> List[Dict[str, Any]]:
        """
        Returns a list of roles the current user can assign to themselves,
        including the scopes in which they can assign them.
        """
        async with self._db_manager.get_db() as db:
            user = await UserManager(self._db_manager).get_by_id(user_id, with_relations=True, db=db)
            if not user:
                return []

            all_roles = await self.get_all_roles()
            final_assignable_roles = []
            user_max_level = -1
            if user.roles:
                user_max_level = max((r.level for r in user.roles if r.level is not None), default=-1)

            for role in all_roles:
                assignable_scopes = set()

                # Pathway 1: Check for specific permission override and get its scopes
                required_permission = f"roles:assign:{role.name}"
                permission_scopes = await self.get_permission_scopes(user.id, required_permission, db)
                for scope in permission_scopes:
                    assignable_scopes.add(scope)

                # Pathways 2 & 3 are generally not scope-specific; they grant the ability to assign globally.
                is_authorized_globally = False

                # Pathway 2: Check for direct role assignment grant
                for user_role in user.roles:
                    if any(assignable.id == role.id for assignable in user_role.can_assign_roles):
                        is_authorized_globally = True
                        break

                if not is_authorized_globally:
                    # Pathway 3: Check role level hierarchy
                    if role.level is not None and user_max_level > role.level:
                        is_authorized_globally = True

                # If authorized globally (by hierarchy or direct grant), add 'global' scope.
                if is_authorized_globally:
                    assignable_scopes.add("global")

                # If we found any scopes for this role, add it to our results.
                if assignable_scopes:
                    final_assignable_roles.append({
                        "name": role.name,
                        "description": role.description,
                        "scopes": sorted(list(assignable_scopes))  # Sort for consistent output
                    })

        return final_assignable_roles

    async def get_users_for_role(self, role_name: str, scope: str = None) -> List[Dict[str, str]]:
        """Fetches all users who have a given role, along with the scope."""
        async with self._db_manager.get_db() as db:
            role = await self.get_by_name(role_name)
            if not role:
                return []
            stmt = select(User.username, user_roles_association.c.scope).join_from(
                user_roles_association, User, user_roles_association.c.user_id == User.id
            ).where(user_roles_association.c.role_id == role.id)
            if scope is not None:
                stmt = stmt.where(
                    user_roles_association.c.scope == scope,
                )
            results = (await db.execute(stmt)).all()
            return [{"username": row[0], "scope": row[1]} for row in results]

    async def _is_authorized_to_manage_role(self, manager_user: User, role_to_manage: Role, db: AsyncSession) -> bool:
        """Private helper to check the 3 authorization pathways for role assignment."""
        # Pathway 1: Permission Override
        required_permission = f"roles:assign:{role_to_manage.name}"
        if await self.has_permission(manager_user.id, required_permission, db=db):
            return True

        # Pathway 2: Direct Grant
        for manager_role in manager_user.roles:
            if any(assignable.id == role_to_manage.id for assignable in manager_role.can_assign_roles):
                return True

        # Pathway 3: Level Hierarchy
        if manager_user.roles:
            manager_max_level = max((role.level for role in manager_user.roles if role.level is not None), default=-1)
            if role_to_manage.level is not None and manager_max_level > role_to_manage.level:
                return True

        return False

    async def assign_to_user(self, user_id: str, role_name: str, assigner_id: str, scope: str = 'none'):
        async with self._db_manager.get_db() as db:
            # Step 1: Fetch all necessary objects
            user_manager = UserManager(self._db_manager)
            assigner = await user_manager.get_by_id(assigner_id, with_relations=True, db=db)
            if not assigner:
                raise UserNotFoundError("Assigner user not found.")

            role_to_assign = await self.get_by_name(role_name)
            if not role_to_assign:
                raise RoleNotFoundError(f"Role '{role_name}' not found.")

            # Step 2: Perform the 3-pathway 'OR' authorization check

            can_manage_role = await self._is_authorized_to_manage_role(assigner, role_to_assign, db)
            if not can_manage_role:
                raise OperationForbiddenError(
                    "You lack the required permission, direct grant, or sufficient role level to assign this role."
                )

            # Step 3: If authorized, proceed with the assignment
            target_user = await user_manager.get_by_id(user_id, db=db)
            if not target_user: raise UserNotFoundError("Target user not found.")

            assoc_stmt = select(user_roles_association).where(
                user_roles_association.c.user_id == user_id,
                user_roles_association.c.role_id == role_to_assign.id,
                user_roles_association.c.scope == scope
            )
            if (await db.execute(assoc_stmt)).first():
                return  # Already assigned

            insert_stmt = user_roles_association.insert().values(
                user_id=target_user.id, role_id=role_to_assign.id, scope=scope,
                given_by_id=assigner_id, given_at=time.time()
            )
            await db.execute(insert_stmt)
            await self._db_manager.log_audit_event(
                user_id, "ROLE_ASSIGNED", "system",
                {"role": role_name, "scope": scope, "by": assigner_id}, db=db
            )
            await db.commit()

    async def remove_from_user(self, user_id: str, role_name: str, remover_id: str, scope: str = 'none'):
        async with self._db_manager.get_db() as db:
            # Step 1: Fetch all necessary objects
            user_manager = UserManager(self._db_manager)
            assigner = await user_manager.get_by_id(remover_id, with_relations=True, db=db)
            if not assigner:
                raise UserNotFoundError("Assigner user not found.")

            role_to_assign = await self.get_by_name(role_name)
            if not role_to_assign:
                raise RoleNotFoundError(f"Role '{role_name}' not found.")

            # Step 2: Perform the 3-pathway 'OR' authorization check

            can_manage_role = self._is_authorized_to_manage_role(assigner, role_to_assign, db)

            if not can_manage_role:
                raise OperationForbiddenError(
                    "You lack the required permission, direct grant, or sufficient role level to assign this role."
                )
            target_user = await user_manager.get_by_id(user_id, db=db)
            if not target_user: raise UserNotFoundError("Target user not found.")
            assoc_stmt = select(user_roles_association).where(
                user_roles_association.c.user_id == user_id,
                user_roles_association.c.role_id == role_to_assign.id,
                user_roles_association.c.scope == scope,
            )
            if (await db.execute(assoc_stmt)).first():
                delete_stmt = user_roles_association.delete().where(
                    user_roles_association.c.user_id == user_id,
                    user_roles_association.c.role_id == role_to_assign.id,
                    user_roles_association.c.scope == scope,
                )
                await db.execute(delete_stmt)
                await self._db_manager.log_audit_event(
                    user_id, "ROLE_REMOVED", "system",
                    {"role": role_name, "scope": scope, "by": remover_id}, db=db
                )
                await db.commit()
            else:
                raise RoleNotFoundError(f"Role '{role_name}' not found for this user.")

    async def get_by_name(self, name: str) -> Optional[Role]:
        async with self._db_manager.get_db() as db:
            stmt = select(Role).where(Role.name == name).options(selectinload(Role.can_assign_roles))
            return (await db.execute(stmt)).unique().scalar_one_or_none()

    async def get_by_id(self, role_id: str) -> Optional[Role]:
        async with self._db_manager.get_db() as db:
            stmt = select(Role).where(Role.id == role_id).options(selectinload(Role.can_assign_roles))
            return (await db.execute(stmt)).unique().scalar_one_or_none()

    async def get_or_create(self, name: str, defaults: dict = None) -> Tuple[Role, bool]:
        role = await self.get_by_name(name)
        if role: return role, False
        new_role = await self.create(name, **(defaults or {}))
        return new_role, True

    async def create(self, name: str, description: str = "", system: bool = False, level: Optional[int] = None) -> Role:
        async with self._db_manager.get_db() as db:
            if await self.get_by_name(name):
                raise ValueError(f"Role with name '{name}' already exists.")
            new_role = Role(name=name, description=description, system=system, level=level)
            db.add(new_role)
            await db.commit()
            await db.refresh(new_role)
            return new_role

    async def add_permission_to_role(self, role_name: str, permission_name: str, adder_id: Optional[str] = None):
        async with self._db_manager.get_db() as db:
            role = await self.get_by_name(role_name)
            if not role: raise RoleNotFoundError(f"Role '{role_name}' not found.")

            perm_stmt = select(Permission).where(Permission.name == permission_name)
            permission = (await db.execute(perm_stmt)).unique().scalar_one_or_none()
            if not permission: raise PermissionNotFoundError(f"Permission '{permission_name}' not found.")

            assoc_stmt = select(role_permissions_association).where(
                role_permissions_association.c.role_id == role.id,
                role_permissions_association.c.permission_id == permission.id
            )
            if (await db.execute(assoc_stmt)).first(): return

            insert_stmt = role_permissions_association.insert().values(
                role_id=role.id, permission_id=permission.id, added_by_id=adder_id, added_at=time.time()
            )
            await db.execute(insert_stmt)
            await db.commit()

    async def grant_relationship(self, granter_role_name: str, grantable_name: str,
                                 grantable_manager, relationship_attr: str, db_override: AsyncSession = None):
        async def _action(db: AsyncSession):
            granter_role = await self.get_by_name(granter_role_name)
            if not granter_role:
                raise RoleNotFoundError(f"Granter role '{granter_role_name}' not found.")

            grantable_item = await grantable_manager.get_by_name(grantable_name)
            if not grantable_item:
                item_type = "Role" if relationship_attr == "can_assign_roles" else "Permission"
                raise RoleNotFoundError(f"{item_type} '{grantable_name}' not found.")

            relationship_list = getattr(granter_role, relationship_attr)
            if grantable_item not in relationship_list:
                relationship_list.append(grantable_item)
                await db.merge(granter_role)
                await db.commit()
            return granter_role, grantable_item

        if db_override:
            return await _action(db_override)
        async with self._db_manager.get_db() as _db:
            return await _action(_db)

    async def get_user_roles_with_scope(self, user_id: str) -> List[dict]:
        async with self._db_manager.get_db() as db:
            stmt = select(Role.name, user_roles_association.c.scope).join_from(
                user_roles_association, Role, user_roles_association.c.role_id == Role.id
            ).where(user_roles_association.c.user_id == user_id)
            results = (await db.execute(stmt)).all()
            return [{"role_name": row[0], "scope": row[1]} for row in results]

    async def has_permission(self, user_id: str, permission_name: str, scope_prefix: Optional[str] = None,
                             db: AsyncSession = None) -> bool:
        async def _check(session: AsyncSession):
            stmt = select(Permission.id).join(
                role_permissions_association, Permission.id == role_permissions_association.c.permission_id
            ).join(
                user_roles_association, role_permissions_association.c.role_id == user_roles_association.c.role_id
            ).where(
                user_roles_association.c.user_id == user_id, Permission.name == permission_name
            )
            if scope_prefix:
                stmt = stmt.where(
                    or_(user_roles_association.c.scope == 'global',
                        user_roles_association.c.scope.startswith(scope_prefix))
                )
            result = (await session.execute(stmt)).first()
            return result is not None

        if db:
            return await _check(db)
        async with self._db_manager.get_db() as db:
            return await _check(db)

    async def revoke_user_role_by_scope(self, user_id: str, role_name: str, scope: str, revoker_id: str,
                                        ip_address: str = "system"):
        """
        Revokes a specific role from a user within a specific scope, with authorization checks.
        """
        async with self._db_manager.get_db() as db:
            # 1. Fetch necessary objects
            user_manager = UserManager(self._db_manager)
            revoker = await user_manager.get_by_id(revoker_id, with_relations=True, db=db)
            if not revoker:
                raise UserNotFoundError("Revoker user not found.")

            role_to_revoke = await self.get_by_name(role_name)
            if not role_to_revoke:
                raise RoleNotFoundError(f"Role '{role_name}' not found.")

            # 2. Authorization Check
            required_permission = f"roles:revoke:{role_name}"
            has_permission_override = await self.has_permission(revoker_id, required_permission, db=db)

            has_sufficient_level = False
            if revoker.roles:
                revoker_max_level = max(role.level for role in revoker.roles if role.level is not None)
                if role_to_revoke.level is not None and revoker_max_level > role_to_revoke.level:
                    has_sufficient_level = True

            if not (has_permission_override or has_sufficient_level):
                raise OperationForbiddenError(
                    "You lack the required permission or a sufficient role level to revoke this role."
                )

            # 3. If authorized, proceed with revocation
            stmt = delete(user_roles_association).where(
                user_roles_association.c.user_id == user_id,
                user_roles_association.c.role_id == role_to_revoke.id,
                user_roles_association.c.scope == scope
            )
            result = await db.execute(stmt)
            await self._db_manager.log_audit_event(
                user_id, "ROLE_REVOKED", ip_address,
                {"role": role_name, "scope": scope, "by": revoker_id}, db=db
            )
            await db.commit()
            return result.rowcount > 0

    async def delete_role(self, role_name: str, deleter_id: str):
        """
        Deletes a role entirely from the system, with authorization checks.
        """
        async with self._db_manager.get_db() as db:
            # Authorization Check: Must have 'admin:manage:roles' permission
            if not await self.has_permission(deleter_id, "admin:manage:roles", db=db):
                raise OperationForbiddenError("You lack the required permission to delete roles.")

            role_to_delete = await self.get_by_name(role_name)
            if not role_to_delete:
                raise RoleNotFoundError(f"Role '{role_name}' not found.")

            # Prevent deletion of system roles
            if role_to_delete.system:
                raise OperationForbiddenError("System roles cannot be deleted.")

            await db.delete(role_to_delete)
            await db.commit()


class PermissionManager:
    def __init__(self, db_manager: DatabaseManager):
        self._db_manager = db_manager

    async def get_by_name(self, name: str) -> Optional[Permission]:
        async with self._db_manager.get_db() as db:
            stmt = select(Permission).where(Permission.name == name)
            return (await db.execute(stmt)).unique().scalar_one_or_none()

    async def get_or_create(self, name: str, defaults: dict = None) -> Tuple[Permission, bool]:
        perm = await self.get_by_name(name)
        if perm: return perm, False
        new_perm = await self.create(name, **(defaults or {}))
        return new_perm, True

    async def create(self, name: str, description: str = "") -> Permission:
        async with self._db_manager.get_db() as db:
            if await self.get_by_name(name):
                raise ValueError(f"Permission with name '{name}' already exists.")
            new_perm = Permission(name=name, description=description)
            db.add(new_perm)
            await db.commit()
            await db.refresh(new_perm)
            return new_perm


class SessionManager:
    def __init__(self, db_manager: DatabaseManager):
        self._db_manager = db_manager

    async def get_by_id(self, session_id: str) -> Optional[DBSession]:
        async with self._db_manager.get_db() as db:
            stmt = select(DBSession).where(DBSession.session_id == session_id)
            result = await db.execute(stmt)
            return result.unique().scalar_one_or_none()

    async def create(self, user_id: str, ip_address: str, region: str, device: str) -> DBSession:
        async with self._db_manager.get_db() as db:
            session = DBSession(
                session_id=encryption_utils.gen_random_string(32), user_id=user_id,
                create_ip=ip_address, last_ip=ip_address, region=region, device=device
            )
            db.add(session)
            await self._db_manager.log_audit_event(user_id, "SESSION_CREATED", ip_address,
                                                   {"device": device, "region": region}, db=db)
            await db.commit()
            await db.refresh(session)
            return session

    async def terminate(self, session_id: str, ip_address: str, errors="ignore"):
        async with self._db_manager.get_db() as db:
            stmt = select(DBSession).where(DBSession.session_id == session_id)
            session = (await db.execute(stmt)).unique().scalar_one_or_none()
            if session:
                await session.terminate(ip_address, self._db_manager, db)
                await db.commit()
                return True
            if errors == "ignore": return False
            raise SessionNotFoundError("Session not found.")

    async def terminate_all_for_user(self, user_id: str, ip_address: str, except_session_id: Optional[str] = None, db: AsyncSession=None):
        async def _ori(_db: AsyncSession):
            stmt = select(DBSession).where(DBSession.user_id == user_id, DBSession.active == True)
            if except_session_id:
                stmt = stmt.where(DBSession.session_id != except_session_id)
            sessions_to_terminate = (await _db.execute(stmt)).scalars().all()
            for session in sessions_to_terminate:
                session.active = False

            await self._db_manager.log_audit_event(user_id, "SESSIONS_TERMINATED_ALL", ip_address,
                                                   {"except": except_session_id}, db=_db)
            await _db.commit()
        if db:
            return await _ori(db)
        else:
            async with self._db_manager.get_db() as db:
                return await _ori(db)


    async def get_all_for_user(self, user_id: str, session_id: str, only_active=True) -> List[DBSession]:
        async with self._db_manager.get_db() as db:
            stmt = select(DBSession).where(DBSession.user_id == user_id)
            if only_active:
                stmt = stmt.where(DBSession.active == True)
            all_sessions = (await db.execute(stmt)).scalars().all()
            for session in all_sessions:
                if session.session_id == session_id:
                    session.active = True
                    break
            return all_sessions


class TokenManager:
    def __init__(self, db_manager: DatabaseManager):
        self._db_manager = db_manager

    async def create(self, user_id: str, purpose: str, expiry_seconds: Optional[int] = None) -> Token:
        async with self._db_manager.get_db() as db:
            expiry = expiry_seconds or settings.TOKENS_EXPIRY_SECONDS
            token = Token(id=encryption_utils.gen_random_string(32), purpose=purpose, user_id=user_id,
                          etime=time.time() + expiry)
            db.add(token)
            await db.commit()
            await db.refresh(token)
            return token

    async def validate(self, db: AsyncSession, token_id: str, purpose: str, ip_address: str) -> User:
        stmt = select(Token).where(Token.id == token_id, Token.purpose == purpose).options(
            joinedload(Token.user).joinedload(User.mfa_methods))
        token_obj = (await db.execute(stmt)).unique().scalar_one_or_none()
        if not token_obj: raise InvalidTokenError("Invalid token.")
        if token_obj.used: raise InvalidTokenError("Token has already been used.")
        user_stmt = select(User).where(User.id == token_obj.user_id)
        user = (await db.execute(user_stmt)).unique().scalar_one_or_none()
        if not user: raise InvalidTokenError("Token is not associated with a valid user.")

        if not token_obj.is_valid():
            new_token = await self.create(user.id, purpose)
            await token_obj.mark_used(ip_address, self._db_manager, db)
            raise TokenExpiredError("Token expired. Reload Page.", new_token_id=new_token.id)

        await token_obj.mark_used(ip_address, self._db_manager, db)
        return user


class AuditManager:
    """Manages querying the audit trail for security and administrative purposes."""

    def __init__(self, db_manager: DatabaseManager):
        self._db_manager = db_manager

    async def get_events_for_user(self, user_id: str, skip: int = 0, limit: int = 25) -> List[AuditEvent]:
        async with self._db_manager.get_db() as db:
            stmt = select(AuditEvent).where(AuditEvent.user_id == user_id).order_by(desc(AuditEvent.timestamp)).offset(
                skip).limit(limit)
            result = await db.execute(stmt)
            return list(result.scalars().all())

    async def get_events_by_type(self, event_type: str, skip: int = 0, limit: int = 100) -> List[AuditEvent]:
        async with self._db_manager.get_db() as db:
            stmt = select(AuditEvent).where(AuditEvent.event_type == event_type).order_by(
                desc(AuditEvent.timestamp)).offset(skip).limit(limit)
            result = await db.execute(stmt)
            return list(result.scalars().all())


class AuthTunaAsync:
    """High-level facade for all authentication and authorization operations."""

    def __init__(self, db_manager: DatabaseManager):
        self.db_manager = db_manager
        self.users = UserManager(db_manager)
        self.roles = RoleManager(db_manager)
        self.permissions = PermissionManager(db_manager)
        self.sessions = SessionManager(db_manager)
        self.tokens = TokenManager(db_manager)
        self.mfa = MFAManager(db_manager)
        self.audit = AuditManager(db_manager)

    async def get_dashboard_stats(self) -> Dict[str, Any]:
        """Gathers all necessary statistics for the admin dashboard."""
        async with self.db_manager.get_db() as db:
            # Get user counts
            total_users_count = await db.scalar(select(func.count(User.id)))
            active_users_count = await db.scalar(select(func.count(User.id)).where(User.is_active == True))
            unverified_users_count = await db.scalar(select(func.count(User.id)).where(User.email_verified == False))

            # Get recent registrations
            recent_users_stmt = select(User).order_by(User.created_at.desc()).limit(5)
            recent_registrations = list((await db.execute(recent_users_stmt)).unique().scalars().all())

            # Get recent login events
            recent_logins_stmt = select(AuditEvent).where(
                or_(AuditEvent.event_type == "LOGIN_SUCCESS", AuditEvent.event_type == "LOGIN_FAILED")
            ).order_by(AuditEvent.timestamp.desc()).limit(10)
            recent_logins = list((await db.execute(recent_logins_stmt)).unique().scalars().all())

            return {
                "total_users": total_users_count,
                "active_users": active_users_count,
                "unverified_users": unverified_users_count,
                "recent_registrations": recent_registrations,
                "recent_logins": recent_logins,
            }

    async def signup(self, username: str, email: str, password: str, ip_address: str) -> Tuple[User, Optional[Token]]:
        user = await self.users.create(
            email=email, username=username, password=password, ip_address=ip_address,
            email_verified=not settings.EMAIL_ENABLED
        )
        token = None
        if settings.EMAIL_ENABLED:
            token = await self.tokens.create(user.id, "email_verification")
        return user, token

    async def login(self, username_or_email: str, password: str, ip_address: str, region: str, device: str) -> Union[
        Tuple[Any, Token], Tuple[Any, Session]]:
        async with self.db_manager.get_db() as db:
            stmt = select(User).where((User.email == username_or_email) | (User.username == username_or_email))
            user = (await db.execute(stmt)).unique().scalar_one_or_none()
            if not user:
                raise InvalidCredentialsError("Incorrect username/email or password.")
            password_valid = await user.check_password(password, ip_address, self.db_manager, db)
            if password_valid is False:
                raise InvalidCredentialsError("Incorrect username/email or password.")
            elif password_valid is None:
                raise EmailNotVerifiedError("Email Not Verified.")
            elif password_valid is True:
                pass
            else:
                raise InvalidCredentialsError("Incorrect username/email or password (config error).")
            if not user.is_active:
                await self.db_manager.log_audit_event(
                    user.id, "LOGIN_FAILED", ip_address, {"reason": "user_suspended"}, db=db
                )
                await db.commit()
                raise OperationForbiddenError("This account has been suspended.")
            await db.commit()
            if user.mfa_enabled:
                return user, await self.tokens.create(user.id, "mfa_validation", expiry_seconds=300)
            else:
                return user, await self.sessions.create(user.id, ip_address, region, device)

    async def change_password(self, user: User, current_password: str, new_password: str, ip_address: str, current_session_id: str):
        """
        Allows a user to change their password after verifying their current one.
        Terminates all other active sessions for security.
        """
        async with self.db_manager.get_db() as db:
            # Re-fetch user inside the session to ensure it's attached
            user_in_session = await db.get(User, user.id)
            if not user_in_session:
                raise UserNotFoundError("User not found.")

            password_valid = await user_in_session.check_password(current_password, ip_address, db_manager_custom=self.db_manager, db=db)
            if not password_valid:
                raise InvalidCredentialsError("The current password you entered is incorrect.")

            await user_in_session.set_password(new_password, ip_address, db_manager_custom=self.db_manager, db=db)
            await self.sessions.terminate_all_for_user(user.id, ip_address, except_session_id=current_session_id, db=db)
            await db.commit()

    async def request_password_reset(self, email: str) -> Optional[Token]:
        async with self.db_manager.get_db() as db:
            user = await self.users.get_by_email(email)
            if not user: return None

            count_stmt = select(func.count()).select_from(Token).where(
                Token.user_id == user.id,
                Token.purpose == "password_reset",
                Token.ctime > time.time() - 86400
            )
            recent_tokens_count = (await db.execute(count_stmt)).scalar()

            if recent_tokens_count >= settings.TOKENS_MAX_COUNT_PER_DAY_PER_USER_PER_ACTION:
                raise RateLimitError("Too many password reset requests.")

            return await self.tokens.create(user.id, "password_reset")

    async def reset_password(self, token_id: str, new_password: str, ip_address: str) -> User:
        """Performs token validation and password reset in a single atomic transaction."""
        async with self.db_manager.get_db() as db:
            user = await self.tokens.validate(db, token_id, "password_reset", ip_address)
            await user.set_password(new_password, ip_address, self.db_manager, db)
            await db.commit()
            return user

    async def verify_email(self, token_id: str, ip_address: str) -> User:
        """Performs email verification in a single atomic transaction."""
        async with self.db_manager.get_db() as db:
            user = await self.tokens.validate(db, token_id, "email_verification", ip_address)
            if not getattr(user, 'email_verified', False):
                user.email_verified = True
            await db.commit()
            return user

    async def validate_mfa_login(self, mfa_token: str, code: str, ip_address: str, device_data: dict) -> DBSession:
        """
        Handles the second step of an MFA login within a single transaction.
        """
        async with self.db_manager.get_db() as db:
            user = await self.tokens.validate(db, mfa_token, "mfa_validation", ip_address)

            user_with_mfa = await self.users.get_by_id(user.id, with_relations=True, db=db)
            if not user_with_mfa or not user_with_mfa.mfa_methods:
                raise InvalidTokenError("MFA is not configured correctly for this user.")

            # Assuming one TOTP method for now
            totp_method = next((m for m in user_with_mfa.mfa_methods if m.method_type == 'totp'), None)
            if not totp_method or not totp_method.secret:
                raise InvalidTokenError("TOTP secret not found.")
            totp = pyotp.TOTP(totp_method.secret)
            is_valid_code = totp.verify(code)
            is_valid_recovery = False
            if not is_valid_code:
                is_valid_recovery = await self.mfa.verify_recovery_code(user, code, db)
            if not is_valid_code and not is_valid_recovery:
                raise InvalidTokenError("Invalid MFA code.")

            await db.commit()
        session = await self.sessions.create(
            user.id, ip_address, device_data["region"], device_data["device"]
        )
        return session



"""
This file gonna contain routes for ui (dashboards, user info and logins etc etc, gonna work on this soon)
"""
from typing import List

from fastapi import APIRouter, Depends, Request, HTTPException
from fastapi.templating import Jinja2Templates
from pydantic import BaseModel
from fastapi import status
from sqlalchemy.exc import IntegrityError

from authtuna.core.config import settings
from authtuna.core.database import User
from authtuna.integrations.fastapi_integration import get_current_user, auth_service

router = APIRouter(prefix="/ui", tags=["ui"])
templates = Jinja2Templates(directory=settings.DASHBOARD_AND_USER_INFO_PAGES_TEMPLATE_DIR)
class UserProfileUpdate(BaseModel):
    username: str

class SessionInfo(BaseModel):
    session_id: str
    region: str
    device: str
    last_ip: str
    mtime: float
    is_current: bool = False

    class Config:
        from_attributes = True

@router.get("/dashboard", name="ui_dashboard")
async def dashboard(request: Request, user: User = Depends(get_current_user)):
    """
    Renders the user dashboard page.
    """
    return templates.TemplateResponse("user_dashboard.html", {"request": request, "user": user})


@router.get("/profile", name="ui_profile")
async def profile(request: Request, user: User = Depends(get_current_user)):
    """
    Renders the user profile page.
    """
    return templates.TemplateResponse("user_profile.html", {"request": request, "user": user})


@router.get("/settings", name="ui_settings")
async def settings_page(request: Request, user: User = Depends(get_current_user)):
    """
    Renders the user settings page.
    """
    return templates.TemplateResponse("settings.html", {"request": request, "user": user})


@router.patch("/profile", status_code=status.HTTP_200_OK)
async def update_profile(
    update_data: UserProfileUpdate,
    request: Request,
    user: User = Depends(get_current_user)
):
    """
    Update the current user's profile.
    """
    try:
        ip_address = request.state.user_ip_address
        updated_user = await auth_service.users.update(
            user_id=user.id,
            update_data={"username": update_data.username},
            ip_address=ip_address
        )
        return {"message": "Profile updated successfully!", "user": {"username": updated_user.username}}
    except IntegrityError:
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail="Username already exists.")
    except Exception as e:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=f"An error occurred: {e}")

@router.get("/settings/sessions", response_model=List[SessionInfo])
async def get_user_sessions(request: Request, user: User = Depends(get_current_user)):
    """
    Fetches all active sessions for the current user.
    """
    current_session_id = request.state.session_id
    sessions = await auth_service.sessions.get_all_for_user(user.id, current_session_id)
    return sessions


@router.post("/settings/sessions/{session_id}/terminate", status_code=status.HTTP_200_OK)
async def terminate_session(session_id: str, request: Request, user: User = Depends(get_current_user)):
    """
    Terminates a specific session for the current user.
    """
    session_to_terminate = await auth_service.sessions.get_by_id(session_id)
    if not session_to_terminate or session_to_terminate.user_id != user.id:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Session not found or does not belong to user.")

    if session_to_terminate.session_id == request.state.session_id:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Cannot terminate the current session.")

    ip_address = request.state.user_ip_address
    await auth_service.sessions.terminate(session_id, ip_address)
    return {"message": "Session terminated successfully."}

@router.post("/settings/sessions/terminate-all", status_code=status.HTTP_200_OK)
async def terminate_all_other_sessions(request: Request, user: User = Depends(get_current_user)):
    """
    Terminates all active sessions for the current user, except the current one.
    """
    current_session_id = request.state.session_id
    ip_address = request.state.user_ip_address
    await auth_service.sessions.terminate_all_for_user(user.id, ip_address, except_session_id=current_session_id)
    return {"message": "All other sessions have been terminated."}
